var OAuth = require('oauth').OAuth,
OAuth2 = require('oauth').OAuth2,
Structr = require('structr'),
vine =require('vine'),
utils = require('sk/node/utils');



var AuthHandler = Structr({
	//the path to the button template
	buttonPath: null,

	'__construct': function()
	{
	},
	
	//authenticates the user and makes sure they're logged in, otherwise re-directing them to a login screen
	'authenticate': function(request,successCallback){},

	//called when redirected to the url after a post-to
	'authorized': function(request,successCallback){},
	
	//checks to see if the authentication keys are correct for a given service such as oauth
	'checkAppKeys': function(keys)
	{
		for(var i in this.appParams)
		{
			if(!keys[i]) return false;
		}
		return true;
	},
	
	'share': function(secret, callback)
	{
		this.getAccountInfo(secret, function(err, userid, info, displayName)
		{
			info = info || {};

			utils.copyTo(secret,info);

			callback(false,{userid:userid,info:info,displayName:displayName});
		})
	}

	//refreshes a connection
	// ,refresh:function(request,successCallback){}

});


var OAuthHandler = AuthHandler.extend({
	appParams:{'consumerKey':{type:'input'},'consumerSecret':{type:'input'}},
	'override __construct': function( authorizeUrl, requestUrl, accessUrl, appParams, version, authorize_callback, signatureMethod, nonceSize, costomHeaders)
	{
		this.authorizeUrl = authorizeUrl;
		this.oauth = new OAuth(requestUrl, accessUrl, appParams.consumerKey, appParams.consumerSecret, version, null, signatureMethod, nonceSize, costomHeaders);
		this._super();
	},

	//authenticates the user of the
	'authenticate': function(request)
	{
		var oa = this.oauth,
		s = this;

		oa._authorize_callback = this.getAuthorizeCallback(request);

		
		oa.getOAuthRequestToken(this.getRequestTokenParams(), function(err, authToken, authTokenSecret, results)
		{
			if(err) console.warn(err);

			/*request.session(function(session)
			{
				session.copy({authToken: authToken, authTokenSecret: authTokenSecret });
				request.redirect(s.authorizeUrl+"?oauth_token="+authToken);
			});*/

			request.session.copy({authToken: authToken, authTokenSecret: authTokenSecret });
			
			request.respond({ redirect: s.authorizeUrl+"?oauth_token="+authToken });

			request.end();
		})  
	},

	'getRequestTokenParams': function()
	{
		return {};
	},

	//called when redirected to the url after a post-to
	'authorized': function(request, callback)
	{
		var oa = this.oauth,
		s = this,
		sess = request.session;
		
		
		var authToken = sess.data.authToken;
		var authTokenSecret = sess.data.authTokenSecret;
		var oauthVerifier = request.data.oauth_verifier;
		
		// sess.remove('authToken','authTokenSecret');

		//if(!authTokenSecret)...

		oa.getOAuthAccessToken(authToken, authTokenSecret, oauthVerifier, function(err, accessToken, accessTokenSecret, result)
		{
			if(err) 
			{
				// console.warn(JSON.stringify(err));
				request.respond({ redirect: 'http://'+request.host+'/connect/'+request.data.service } );
				return request.end();
			}

			result.accessToken = accessToken;
			result.accessTokenSecret = accessTokenSecret;
			
			
			s.share({accessToken: accessToken, accessTokenSecret: accessTokenSecret }, function(err, info)
			{
				callback(false, info);
			});

		});  
		
	},
	'getAuthorizeCallback': function(request)
	{
		return null;
	},
	'getAccountInfo': function(result,callback)
	{
		//abstract
	},
	
	//FUCKING OAUTH API doesn't have a propper request method, so this is here AENKLFNSDLFKGBDSLFKBNDS
	'_request': function(url,method,account,data,callback)
	{
		var tok = account.secret.accessToken,
		sec = account.secret.accessTokenSecret;
		 
		if(method == 'PUT' || method == 'POST') return this.oauth._putOrPost(method,url,tok,sec,data,callback);
		                     
		this.oauth._performSecureRequest(tok,sec,method,url,data,null,null,callback);
	},
	'get': function(url,account,data,callback)
	{
		this._request(url,"GET",account,data,callback);
	},
	'post': function(url,account,data,callback)
	{
		this._request(url,"POST",account,data,callback);
	},
	'remove': function(url,account,data,callback)
	{
		this._request(url,"DELETE",account,data,callback);
	},
});


var OAuth2Handler = AuthHandler.extend({
	
	'appParams': {'clientId':{type:'input'},'clientSecret':{type:'input'}},
	
	'override __construct': function( appParams, baseSite, authorizePath, accessTokenPath)
	{

		this.oauth = new OAuth2(appParams.clientId, appParams.clientSecret, baseSite, authorizePath, accessTokenPath);

		this._super()
	},
	
	//authenticates the user of the
	'authenticate': function(request)
	{
		var oa = this.oauth,
		s = this;

		var url = oa.getAuthorizeUrl(this.getParams(request));

		// request.redirect(url);

		request.respond({ redirect: url });
		request.end();
	},
	'getParams': function(request)
	{
		return {}
	},

	//called when redirected to the url after a post-to
	'authorized': function(request, callback)
	{
		var oa = this.oauth,
		s = this;
		
		oa.getOAuthAccessToken(request.data.code, this.getParams(request), function(err, accessToken, refreshToken)
		{
			if(err) return vine.api().error("Cannot add account. Please try again later.").end(request);

			var tokenParts = accessToken.split(accessToken);
			
			s.share({ accessToken: accessToken }, function(err, data)
			{
				callback(false, data);
			})

			s.updateAccount(accessToken);

		});
	},
	'updateAccount':function(accessToken, callback){}
});



/*var AuthBasicHandler = AuthHandler.extend({
	name:'basic',
	'__construct': function()
	{
		this._super();

		var s = this;
		this.basic = new BasicAuth(function(user,password,callback)
		{
			s.delegate.getProfile(user,{password:password},callback);
		});
	},
	'authenticate': function(request,callback)
	{
		this.basic.authenticate(request, callback);  
	}
});*/


exports.AuthHandler		 = AuthHandler;
exports.OAuthHandler     = OAuthHandler;
exports.OAuth2Handler    = OAuth2Handler;
// exports.AuthBasicHandler = AuthBasicHandler;



