var utils = require('../utils'),
	EventEmitter = require('events').EventEmitter,
	ByteArray = require('./bytes').ByteArray;
	
	
function returnSimpleData(data)
{
	return data;
}
function returnNull()
{
	return null;
}

function defaultFactory(type)
{
	
	
	var handler = null;
	
	switch(type)
	{
		case 0: handler =  
		{
			readExternal:function(input)
			{
				return input.readFloat();
			},
			toJSON:returnSimpleData,
			fromJSON:returnSimpleData
		}
		break;//NUMBER/float
		case 1: handler = 
		{
			readExternal:function(input)
			{
				return input.readInt();
			},
			toJSON:returnSimpleData,
			fromJSON:returnSimpleData
		}
		break;//INT
		case 2: handler =
		{
			readExternal:function(input)
			{
				return input.readInt();
			},
			toJSON:returnSimpleData,
			fromJSON:returnSimpleData
		}
		break;//UINT
		case 3: handler = 
		{
			readExternal:function(input)
			{
				return input.readUTFBytes(input.readInt());
			},
			toJSON:returnSimpleData,
			fromJSON:returnSimpleData
		}
		break;//STRING
		case 4: handler = 
		{
			readExternal:function(input)
			{
				return input.readBoolean();
			},
			toJSON:returnSimpleData,
			fromJSON:returnSimpleData
		}
		break;//BOOLEAN
		case 6: handler = 
		{
			readExternal:function(input)
			{
				if(input.readBoolean())
				{
					var ba = new ByteArray();
					var n = input.readInt();
					if(n) input.readBytes(ba,0,n);
					return ba;
				}
				
				return null;
			},
			toJSON:returnNull,
			fromJSON:returnNull
		}
		break;//BYTEARRAY
		default: handler = 
		{
			readExternal:returnNull,
			toJSON:returnNull,
			fromJSON:returnNull
		}
		break;
	}
	return handler;
	
}

function getDataType(data)
{
	switch(typeof(data))
	{
		case "number": return 0;
		case "string": return 3;
		case "boolean": return 4;
		default: return data ? data.type : -1;
	}
}

var Setting = utils.extend({},
{
	bytes:null,
	__construct:function(factory,type)
	{
		this._factory = factory;
		this.type = type;
		this.emitter = new EventEmitter();
	},
	on:function(type,callback)
	{
		this.emitter.addListener(type,callback);
	},
	getData:function()
	{
		if(this.bytes)
		{
			var ba = this.bytes;
			this.bytes = null;
			this._data = this._factory(this.type).readExternal(ba);
		}
		
		return this._data;
	},
	setData:function(value)
	{
		this._data = value;
		this.emitter.emit('change',value);
	},
	dispose:function()
	{
		this.emitter.removeListeners('change');
	},
	writeExternal:function(output)
	{
		var type = this.type;
		
		//if the type doesn't exist then determine what it is. if it's a custom data type, then it must be
		//attached to the object
		if(!type)
		{
			type = getDataType(this._data);
		}
		
		this._factory(type).writeExternal(this._data,output);
	},
	toJSON:function()
	{
		return this._factory(this.type).toJSON(this.getData());
	},
	fromJSON:function(data)
	{
		this.type = getDataType(data);
		this._data = this._factory(this.type).fromJSON(data);
	},
	readExternal:function(input)
	{
		this.bytes = input;
	}
});


var SettingTable = utils.extend({},
{
	__construct:function(factory)
	{
		this._length = 0;
		this._factory = factory || defaultFactory;
		this._settings = {};
	},
	getSetting:function(name)
	{
		if(this._settings[name])
		{
			return this._settings[name];
		}
		
		var setting = new Setting(this._factory);
		this._settings[name] = setting;
		this._length++;
		
		setting.on('change',function()
		{
			
		});
		
		return this._settings[name];
	},
	removeSetting:function(name)
	{
		var setting = this._settings[name];
		if(setting) setting.dispose();
		delete this._settings[name];
	},
	removeAlLSettings:function()
	{
		for(var i in this._settings)
		{
			this.removeSetting(i);
		}
	},
	hasSetting:function(name)
	{
		return this._settings[name] != undefined;
	},
	readExternal:function(input)
	{
		input.readInt();//version
		
		var n = input.readInt(),//number if items
			ba,
			i = 0,
			j,
			type,
			name,
			existing,
			setting;
		   
				
		while(i < n)
		{
			ba = new ByteArray();
			
			type = input.readInt();
			name = input.readUTFBytes(input.readInt());
			
			
			if(!this._settings[name])
			{
				setting = this.getSetting(name);
			}else
			{
				setting= this._settings[name];
			}
			setting.type = type;
			
			j = input.readInt();
			
			
			if(j) 
			{
				input.readBytes(ba,0,j);
			}
			
			
			setting.readExternal(ba);
			i++;
			
			// ba.clear();
		}
		
	},
	toJSON:function()
	{
		var data = {};
		
		for(var i in this._settings)
		{
			data[i] = this._settings[i].toJSON();
		}
		
		return data;
	},
	fromSettingTable:function(settingTable)
	{
		for(var i in settingTable._settings)
		{
			if(this._settings[i])
			{
				this._settings[i]._data = settingTable._settings[i]._data;
			}
		}
	},
	fromJSON:function(data)
	{
		
		for(var i in data)
		{
			var setting = new Setting(this._factory);
			setting.fromJSON(data[i]);
			this._settings[i] = setting;
		}
	},
	writeExternal:function(output)
	{
		/*
		$output->writeInt(1);//version
		$output->writeInt($this->_length);

		$ba = new ByteArray();
		
		foreach($this->_settings as $k => $setting)
		{	                   
			            
			
			
			$ba->clear();
			
			$output->writeInt($setting->getType());
			$output->writeInt(strlen($k));
			$output->writeUTFBytes($k);
			$setting->writeExternal($ba);
			$output->writeInt($ba->getLength());
			$output->writeBytes($ba);
			
			
		}
		*/
	}
	
}); 

function peek(input,n,o)
{
	
	// input.pos-=100;
	// input.pos-=100;
	var buffer = input.peek(n||30,o);
	// input.pos+=100;s
	for(var k = 0; k < buffer.length; k++)
	{
		console.log(buffer.charCodeAt(k));
	}
}

exports.peek = peek;
exports.SettingTable = SettingTable;
exports.defaultFactory = defaultFactory;