var htmlparser = require('htmlparser'),
	events = require('events'),
	net = require('../io/net'),
	sys = require('sys'),
	URLRequest = net.Request,
	utils = require('sk/node/utils'),
	loadRequest = net.loadRequest,
	BatchQueue = require('sk/core/queue').BatchQueue,
	Url = require('url');
 

//some quick notes about what I've noticed when developing this script, and how it's grown to where it's at:
//image patterns-----------------------------------------------------------------------------------------------
//1. images sometimes are in dated folders e.g: 2011/03/my-image.jpeg   - positive
//2. images have the title of the article in their name lots of times - SUPER positive
//3. images in articles don't typically have anchor links wrapped around them. ones that do typically link to the same article. (probably backend related) - SUPER negative
//4. good images are typically pngs, and jpegs. gifs are usually reserved for graphical elements - negative
//5. lots of times images have the width/height specified in their name e.g: image-100x150.jpg - neutral 
//6. sometimes images provide "large" and "small" in their name e.g: my-image_sm.jpg, my-image_l.jpg - neutral
//7. GOOD names typically nave more than just a-z characters. if am image JUST has a-z chars, then it's most likely a graphical element.


//titles
//1. have h tags

//text-------------------------------------------------------------------------------------------------
                
var regexp = 
{
	unlikelyCandidates:    /combx|comment|community|disqus|extra|foot|header|menu|remark|rss|shoutbox|sidebar|sponsor|ad-break|agegate|pagination|pager|popup|tweet|twitter/i
    ,okMaybeItsACandidate:  /and|article|body|column|main|shadow|hentry|post/i
	,positive:/article|body|content|entry|hentry|main|page|pagination|post|text|blog|story/i
    ,negative:              /combx|comment|com-|contact|foot|footer|footnote|masthead|media|meta|outbrain|promo|related|scroll|shoutbox|sidebar|sponsor|shopping|tags|tool|widget|captcha/i
    ,divToPElements:        /(a|blockquote|dl|div|img|ol|p|pre|table|ul)/i
	,articleParts:/td|p|pre/i //img|object
    ,replaceBrs:            /(<br[^>]*>[ \n\r\t]*){2,}/gi

	//images
	//NOTE: careful with "ads", also in words such as uploads
	,nowayImageName: /([^a-z](share|header|favicon|theme|widget|loader)[^a-z])|promo|([^a-z](ads|advert))|icon|plugin|transparent|footer|(side(.)?bar)/i
	,negativeImageName: /thumb|square|tiny|pixel|social|profile|free|home|picture|small|twitter|comment|account|login|signup|brand|logo|background|bg|mini|small|ajax|banner|template|avatar/gi
	,positiveImageName: /large|upload/i
	,largeImageSize: /([-_](n|c|lg)\.\w+$)/i
	,smallImageSize: /([-_](s|t)\w*\.\w+$)/i
	,positiveUrlCheck: /((youtube|vimeo)\.|slideshow)/i
	,nagativeImageFormat: /\.(gif|ico)/
	,imageDimensions: /(\d+)[x*](\d+)/i
	,datedFolder: /\/\d{2,4}\/\d{2,4}/i
	,acceptableFormats: /jpeg|gif|png|jpg|tiff/i
	
	
	,titleNonWords:/[^\w\s]/g
	
	//videos
	,positiveVideoCheck:/youtube|vimeo|mtvnservices|brightcove/i
	
    ,replaceFonts:          /<(\/?)font[^>]*>/gi
	,punctuation:/[\uff0c,!\.\?']+/gi          
	
	//cleaning
	,cleanRemoveElements:/td|tr/ig
}



	tagScores = {};
	

function setTagScores(stack,score)
{
	stack.forEach(function(value)
	{
		tagScores[value] = score;
	})
}
setTagScores(['div'],5);
setTagScores(['pre','td','blockquote'],3);
setTagScores(['address','ol','ul','dl','dd','dt','li','form'],-3);
setTagScores(['h1','h2','h3','h4','h5','h6','th'],-5);



var Parser = function()
{
	var _self = this,
		_parser,
		_currentUrl = null,
		SKIP_UNLIKELY_CANDIDATES = true,
		metadataHandlers = [];
	
	this.load = function(url,callback)
	{
		var request = new URLRequest(url);
		
		loadRequest(request,function(data)
		{
			_self.parse(request.url,data,callback);
		})
	}
	
	this.addMetadataHandler = function(handler)
	{
		metadataHandlers.push(handler);
	}
	
	this.parse = function(url,htmlText,callback)
	{
		function handleParsedDOM(err,dom)
		{
			SKIP_UNLIKELY_CANDIDATES = true;
			
			var easierToRead = {},
				body = getBody(dom),
				title = getTitle(dom),
				ops = {},
				urlParts = url ? new URLRequest(url) : null;
				
			if(urlParts) urlParts.build();
			
			if(!body || err) return callback({error:true,link:url});
			
			
			//htmlparser doesn't have parent / child relationships - we need it
			setParentChildLink(body);
		
			if(title && title.children)
			{
				title = title.children[0].data;
				
				
				
				//join the title, and remove any punctuation before, and after the article title
				//also remove any HTML tags. This happens sometimes >.>
				// title = title;//titleParts.join(' ').replace(/^[\s-.?!|]*(.*?)[\s-.?!|]*$/,'$1').replace(/<.*?>/g,'')
				
				var titleParts = fixArticleTitle(title, urlParts);
				
				if(titleParts.words.length)
				{
					ops.title = getArticleTitle(body,titleParts) || titleParts;
					
				}
				
			}
			
			ops.location = urlParts;
				
			
			
			//build it.
			var articleData = buildArticle(body,ops);          
			
			_disectArticle(articleData,ops,function(data)
			{
				if(ops.title) data.title = ops.title.value
				data.link = url;
				                                
				                            
				//we want the title of the HTML document, not the article for the UID, since there is a chance we may get the wrong title, and
				//end up with some generic header. the title of the html document is probably our best bet, unless it's dynamic Ugh >.>. the title ALSO
				//might not exist. in THAT case, we need to use the url. The url's not used originally since it may contain getter info (referals to RSS and shit),
				var uid = titleParts ? urlParts.hostname+titleParts.value : url;   
				
				data.uid = data.title ? (uid).replace(/ /g,'') : urlParts.url;
				
				callback(data);
			})
		}
		
		
		
		var handler = new htmlparser.DefaultHandler(handleParsedDOM,{verbose:false,ignoreWhitespace:true});
		_parser = new htmlparser.Parser(handler);
		
		_currentUrl = url;
		
		//trim the whitespace up.
		_parser.parseComplete((htmlText || '').replace(/\s+/g," "));
	}
	
	function parseHandler(err,dom)
	{
		if(err) return false;
	}
	
	function _disectArticle(article,ops,callback,findSize)
	{
		findSize = findSize == undefined ? true : findSize;
		
		//queue up so we can fetch metadata
		var imageMetadata = [], 
		videoMetadata = [],
		otherMetadata = [],
		media = article.media,
		usedImages = {};
		
		
		media.forEach(function(media)
		{
			if(media.type == 'image')
			{
				var node = media.child;
				var src = media.src;
				
                        
				
				src = fixUrl(src,ops);

				if(!usedImages[src])    
				{
					usedImages[src] = src;                                                                        
					var imgSize = media.width+media.height;
					imageMetadata.unshift({type:'image',
											link:src,
											score:media.score,
											distance:media.distance,
											regular:{
												link:src,
												width:media.width,
												height:media.height,
												}});
				}
			}
			else
			if(media.type == 'video')
			{
				videoMetadata.push({type:'video',html:media.child.outerHTML,score:media.score,distance:media.distance});
			}
			else
			{
				otherMetadata.push(media);
			}
		});





		function sortMedia(stack)
		{            
			//sort images if any siblings are a size difference of 20%
			stack.sort(function(a,b)
			{
				return b.score - a.score;
				// return (b.score+b.sortSize)-(a.score+a.sortSize);// - b.score*b.size;   
			})
		}      

		function sortMediaByDistance(stack)
		{
			stack.sort(function(a,b)
			{
				return a.distance - b.distance;
			})
		}               

		function finish()
		{
			// article.url = window.location.url;

			//sort so we can splice with the top images
			sortMedia(imageMetadata);      
			
			
			//tack on the videos.
			var metadata = videoMetadata.concat(otherMetadata.concat(imageMetadata.splice(0,5)));//keep the first 5 images. 
			
			//resort so the videos are in the right spot
			if(videoMetadata.length) sortMedia(metadata);
			
			if(article.metadata)
			{
				metadata = metadata.concat(article.metadata)
			}
			
			callback({content:article.content,metadata:metadata});  

		}     

		//we want the CLOSEST image            


		// sortMediaByDistance(imageMetadata); 
		// var closestImage = imageMetadata[0];

		var toGetSize = [];

		if(findSize && false)
		{
			for(var i = imageMetadata.length; i--;)
			{
				if(imageMetadata[i].size == 0) toGetSize.push(imageMetadata[i])
			}
		}      

		//if the largest image size is small, then we need to check any image without a size, what size they really are...
		//Note: this isn't TOO common, so it might not be an issue if we look at all the images. Also, from the tests I've done, the max
		//is right around 11 or so...
		if(toGetSize.length)
		{   

			//only the first 5 images
			// toGetSize = toGetSize.splice(0,Math.min(toGetSize.length,10));

			var numLoaded = toGetSize.length;

			console.ok('check content size for '+numLoaded+' images.');

			var minImageSizes = 
			{
				'jpeg':2
				,'jpg':2
				,'gif':2
				,'png':10
			}

			//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
			//TODO - if an image is spliced from the metadata, then check the FOLLOWING 
			//iamges to make sure they're:
			//1. Not within the same folder structure
			//2. Not *NEAR* each other - maybe if(Math.abs(prev.distance-next.distance) < 10)
			//3. if the extension is the same
			//if the conditions are met, then dispose the following items before 
			//!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


			var q = new BatchQueue(true,5),
				badFolders = [],
				badDistance = 0


				function next()
				{
					if(!(--numLoaded))
					{
						finish();
					}

					q.next();
				}    


			toGetSize.forEach(function(metadata,index)
			{
				//this is a lot slower.
				q.add(function()
				{	                       
					var folders = metadata.link.replace(/\w+:\/\/[^\/]+\//,'').split('/'),
						imageName = folders.pop(),
						foldersStr = folders.join('/');

					var distance = Math.abs(badDistance - metadata.distance);

					if(badFolders.indexOf(foldersStr) > -1 && distance < 10)
					{                                   
						imageMetadata.splice(imageMetadata.indexOf(metadata),1);
						return next();
					}


					net.getContentLength(metadata.link,function(sizeBytes,type)
					{
						sizeBytes = metadata.size = Number(sizeBytes);
						metadata.sizeType = metadata.size > 0 ? 'b' : 'n';


						var sizeKB = sizeBytes/1024,
							imageType = (type || '').split('/').pop(),
							dumpImage = imageType.search(regexp.acceptableFormats) == -1;

						if(sizeKB < 6 && metadata.distance > 15)
						{
							metadata.score -= 20;

							//okay, it's a tracking pixel.
							dumpImage = dumpImage || imageType == 'gif';
						}	
						else
						{
							//this needs to stay low because elements like logos, and banners can greatly
							//how the image is sorted if this is HIGH.
							metadata.score += sizeBytes/200/metadata.distance;
						}

						//maybe check if the 
						dumpImage = dumpImage || sizeKB != 0 && sizeKB < (minImageSizes[imageType] || 10);

						if(dumpImage)
						{                         
							//don't add anything with folder that look like dates. These could be posts

							//look for 2011/03 or 03/2011   
							if(!foldersStr.match(/\/\d{2,4}\/\d{2,4}/))  
							{
								badFolders.push(foldersStr);
								badDistance = metadata.distance;                                           
							}

							imageMetadata.splice(imageMetadata.indexOf(metadata),1);
						}

						return next();

					})
				});	
			})

			// toGetSize.forEach(function(metadata,index)
			// 		{
			// 			//get the file size of the image
			// 			net.getContentLength(metadata.link,function(sizeBytes,type)
			// 			{
			// 
			// 				var imageType = (type || '').split('/').pop().toLowerCase();
			// 
			// 				var minSize = minImageSizes[imageType] || 10;
			// 				//we set the minimum size for the image based on file type, since different
			// 				//file types compress images differently.
			// 
			// 				var sizeKB = Number(sizeBytes)/1024,
			// 					dumpImage = type == undefined;
			// 					
			// 				metadata.size = sizeBytes;//Number(sizeKB)*10;
			// 				
			// 				
			// 				//we want small images, regardless of where they are to gtfo the top. 
			// 				if(sizeKB < 10)
			// 				{
			// 					metadata.score -= 20;
			// 					
			// 					//okay, it's a tracking pixel.
			// 					dumpImage = dumpImage || imageType == 'gif';
			// 				}
			// 				else
			// 				{
			// 
			// 					//this needs to stay low because elements like logos, and banners can greatly
			// 					//how the image is sorted if this is HIGH.
			// 					metadata.score += sizeBytes/100/metadata.distance;
			// 				}
			// 
			// 				// metadata.sortSize = sizeKB/10; 
			// 				metadata.sizeType = 'b';  
			// 
			// 
			// 
			// 				//content length might NOT come back
			// 				//if the image is small, then push it to the end
			// 				if(dumpImage || (sizeBytes != 0 && sizeKB < minSize))
			// 				{                        
			// 					imageMetadata.splice(imageMetadata.indexOf(metadata),1);
			// 
			// 					//if the image is greater than 7KB push the image at the end of the list - don't delete it.
			// 					if(sizeKB > 7)
			// 					{
			// 						imageMetadata.push(metadata)
			// 					}
			// 				}         
			// 
			// 
			// 				next();
			// 			})  
			// 			
			// 		})

		}
		else
		{
			finish();
		}

	}
	
	
	function findRichestNode(children)
	{
		
		var richest = null;
		
		for(var i = children.length; i--;)
		{
			var node = children[i];
			var score = node.type == 'text' ? (node.data || '').split(regexp.punctuation).length : 0;
			
			var deepest = {node:node,score:score};
			
			if(!richest || score > richest.score) richest = deepest;
			
			if(node.children)
			{
				var richestChild = findRichestNode(node.children);
				if(richestChild && richestChild.score > richest.score) richest = richestChild;
			}
		}
		
		
		return richest ? richest : null;
	}
	
	function getImageData(child, distanceFromArticle, ops)
	{
		
		var score = 100,
			attrs = child.attribs;

		var src = attrs ? (attrs['data-src'] || attrs.src) : null;	
		// var src = attrs ? attrs.src : null;
		
		if(src == '' || !src) return null; 
		  
		var target = child.parent,
			tattrs = target ? target.attribs : null,
			titleMatch = null,
			imageName = src ? src.split('/').pop() : null;
		
		//if the image has a link attached, it's probably not part of the article, but we
		//check anyways incase it is...
		
		if(target && target.name == 'a')
		{          
			var href = tattrs ? tattrs.href : null,
				hrefName = href ? href.split('/').pop() : null;
			
			
			//shit we DEFINITELY don't want. NO href probably means it's a javascript function
			if(!href || href == '')
			{
				return null;
			}
			else
			
			//does the link reference a link such as youtube, or vimeo? slideshow?
			if(href.search(regexp.positiveUrlCheck) > -1)
			{                
				score += 20;
			}
				else                                               

				//does the link execute javascript?
				if(  
					
					//this COULD be an action button
					//href.substr(0,1) != '#' &&
				//reduntant - checked below if there's a linked image                   
				
				//grab only the word part of the name, since it could be a small image linking to a larger image...
				hrefName.split(/\W/).shift() != imageName.split(/\W/).shift() &&

				//does the image link to the article?
				// self.request.url.indexOf(href) == -1 &&    

				//does the link have parts of the article title in it?  
				//TODO: only match the last parts of the uri                   
				(ops.title && href.search(ops.title.regex) == -1) &&               

				//does the link go to another image?
				href.search(/jpg|png|tif|psd|tif|jpeg|gif/i) == -1)
				{                          
					              
					//if the image is close enough to the content, then let's keep it.  
					//this is useful for pages WITHOUT an article. 
					if(distanceFromArticle < 30 && src.search(regexp.nagativeImageFormat) == -1)     // < 30 = find best
					{                    
						score -= 30;
					}
					else
					{                     
						return null;
					}
				}
			}        
			
			if(src.search(regexp.nowayImageName) > -1) return null;
			
			var matches,
			divider = 1,
			imageWithoutExt = imageName.split('.').shift();

			
				
			//title in article? we'll add it, only if it looks human generated. It's NOT human generated if there's a bunch of other shit
			if(ops.title && (titleMatch || (titleMatch = imageName.match(ops.title.regex))) && imageName.match(/[^a-z0-9]/g).length < 20)
			{
				//30 screws up http://www.huffingtonpost.com/2011/03/12/japan-nuclear-plant-explosion_n_834867.html
				score += (titleMatch ||'').length*10;
			}
			


			if(matches = src.match(regexp.negativeImageName))
			{
				score -= matches.length*10; 
			}

			if(imageName.search(regexp.nagativeImageFormat) > -1)
			{
				score -= 10;
			}
			
			//good names have additional text. we'll assume images with generic names are graphical elements
			//also make sure there's no getter data attached (server related). this is common especially with services such as yfrog, and flickr
			if(imageWithoutExt.search(/[^a-z]/) == -1 && imageName.indexOf('?') == -1)
			{
				score -= 30;
			}
			

			/*if(imageName.length < 10)
			{
			score -= 20;
			}*/

			/*if(imageName.search(/\d+/) == -1)
			{
			score -= 10;
			}*/    
			

			if(src.search(regexp.positiveImageName) > -1)
			{
				score += 10;
			}

			if(src.search(regexp.largeImageSize) > -1)
			{
				score += 20;
			}
			else
			if(src.search(regexp.smallImageSize) > -1)
			{
				score -= 20;
			}
			
			if(src.search(regexp.datedFolder) > -1)
			{
				score += 20;
			}

			/*if(src.indexOf("?") > -1)
			{
				score -= 10;
			}*/
			


			// score += src.length/10;


			//use this?
			// src.search(/\?/) > -1



			/*if(articleTitleRegex)
			{
			//only grab the file name.
			var match = src.split('/').pop().match(articleTitleRegex);
			if(match)
			{
			imageScore = imageScore * (1+match.length)
			}
			}*/

			var width  = Number(attrs ? attrs.width || 0 : 0),
			height = Number(attrs ? attrs.height || 0 : 0);   

			if(!width || !height)
			{
				//look for the size in the header
				var dimensions = src.match(regexp.imageDimensions);

				if(dimensions)
				{
					width = Number(dimensions[1]);
					height = Number(dimensions[2]);
					child.width = width;
					child.height = height;
				}
			}

			var sizeScore = 1;

			if(width || height)
			{
				var size = Number(width)+Number(height),
				ratio = Math.min(width,height)/Math.max(width,height);

				//for now, let's see how well it parses. Some websites set the width and height
				//of an image, when they shouldn't >.>

				// if((width < 100 || height < 100) && ratio < .3) continue;


				if((width && height && size < 100)     || 
				(width && width < 75)   ||
				(height && height < 75) ||
				(width && height && ratio < .2))
				{                     

					if(distanceFromArticle > 30) return null;
					// if(distanceFromArticle > 15 || ratio < .2) continue;        
					if(distanceFromArticle > 15) score -= 30;
				}
				else
				{                
					//a image more square, and larger size gets a better score. We use a ratio as a factor
					//because we need to rule out items such as banners. This score however is such a small portion in weighing
					//image since size alone shouldn't determine whether we use the image. 
					// sizeScore = (ratio * size)/10;
					//skinny images are weighed less than 

					// score += sizeScore/10;

					//closer the image is to the article, the more size makes a difference.
					score += size/2/distanceFromArticle;
					
				}

			}	   	                

			var finalScore = score-distanceFromArticle;   
			
			return finalScore < -20 ? null :{type:'image',child:child,width:width || 0,height:height || 0,score:finalScore,src:src,distance:distanceFromArticle}      

		}
	
		function isUnlikely(node)
		{
			var attrs = node.attribs || {};
				
			var uid = String(attrs['class'] + attrs.id);
			
			if(uid && uid.search(regexp.unlikelyCandidates)  !== -1 &&
			   uid.search(regexp.okMaybeItsACandidate) == -1)
			{
				// node.parent
				return true;
			}
			return false;
		}
	
		var getBody = _findNodePartial('body');
		var getTitle = _findNodePartial('title');
		var findArticleTitle = function(children, pageTitle, candidates)
		{
			var child,
				matches;
				
			if(children)
			for(var i = children.length; i--;)
			{
				(child = children[i]);
				if(junkNodes[child.name]) continue;
				
				if((child = children[i]).data &&
					child.data.length < 250 && 
					(matches = child.data.match(pageTitle.regex)) && 
					(matches.length / pageTitle.words.length) > .3 &&     
					                                                   
					//no shit please. 
					child.data.search(/[^\w"?.$%<>]/) == -1)
				{
					
					//take the number of items and divide it by the difference between the title, and text length. We want the closest item possible. Also
					//100 is pulled out of my ass.             
					        
					
					child.titleScore = (matches.length*100)/Math.max(1,Math.abs(pageTitle.regex.source.length-child.data.length));
					
					//again, another random number, but hN tags are a GOOD thing
					if(child.parent.name.match(/^h\d/)) child.titleScore *= 10;
					
					
					candidates.push(child);
				}
				else if(!junkNodes[child.name] && child.children && (child = findArticleTitle(child.children,pageTitle,candidates))) return child;
			}
		}
		
		function fixArticleTitle(title, location)
		{
			if(!title) title = '';

			//build some regexp so we can find the REAL title in the article. This will also help us find images which ALSO have the article
			//in their name
			
			//now we have the parts, lets rebuild it.
			var titleParts = title.toLowerCase().split(' '),titleWords = [],
				usedWords = {};
				

			//first, remove anything in the title that's distracting
			for(var i = titleParts.length; i--;)
			{
				var tp = titleParts[i];

				//remove anything with the domain in it

				//remove any punctuation from the search
				var toSearch = tp.split(/\W/).shift().toLowerCase();
				if(toSearch != '' && location && location.hostname.indexOf(toSearch) > -1)
				{
					titleParts.splice(i,1);
				}
				else
				if(tp.match(/[a-z]/))
				{
					var word = tp.match(/\w+/)[0];
					
					//skip any words that might be a bit generic. e.g: A,me,I,and. we want just the good shit.
					if(word.length > 3 && !usedWords[word])
					{
						usedWords[word] = 1;
						titleWords.push(word);
					}
				}
			}	
			
			//remove non ascii characters. we do this because gunzipped data doesn't play nice with unicode characters
			return {value:title.replace(/&[^;]+;/g,' ').replace(/[^\x20-\x7E]+/g,''),punctLength:title.split(regexp.titleNonWords).length,regex:new RegExp(titleWords.join('|'),'ig'),words:titleWords};
		}
		
		function getArticleTitle(body,pageTitle)
		{
			var candidates = [];
			findArticleTitle(body.children,pageTitle,candidates);
			
			var topCandidate;
			
			for(var i = candidates.length; i--;)
			{
				var candidate = candidates[i];
				
				if(!topCandidate || candidate.titleScore > topCandidate.titleScore)
				{
					topCandidate = candidate;
				}
			}
			
			return topCandidate ? fixArticleTitle(topCandidate.data) : null;
		}
		
		function _findNodePartial(test)
		{
			if(typeof test == 'string')
			{
				var name = test;
				test = function(node)
				{
					return node.name == name;
				}
			}
			
			var func = function(children)
			{
				var child;
				for(var i = children.length; i--;)
				{
					if(test(child = children[i]))
					{
						return child;
					}
					else if(child.children && (child = func(child.children))) return child;
				}
			}
			
			return func;
		}
		
		var junkNodes = {script:1,noscript:1,form:1,select:1,input:1,style:1},
			keepNodes = {b:1,i:1,strong:1,em:1};
		
		
		
		function findArticleCandidates(node,nodesToScore)
		{
			if(!node) return;
			
			if(node.type == 'comment' || junkNodes[node.name])
			{
				node.parent.children.splice(node.parent.children.indexOf(node),1);
				return;
			};
			//no text elements
			if(node.type == 'text')
			{
				return node.textContent = ' '+node.data;
			}
			
			var attrs = node.attribs || {},
				nodeName = node.name ;
				
			if(SKIP_UNLIKELY_CANDIDATES)
			{
				if(isUnlikely(node) && nodeName != 'body')
				{
					node.parent.children.splice(node.parent.children.indexOf(node),1);
					return '';
				}
			}	
			
			
			
			if(nodeName.search(regexp.articleParts) > -1)
			{
				nodesToScore.push(node);
			}
			
			var children = node.children,
				textContent = '',
				childTags = '';
			
			if(children)
			{
				for(var i = children.length; i--;)
				{
					var child = children[i];
					child.parent = node;
					textContent = findArticleCandidates(child,nodesToScore)+textContent;
					if(child.name) childTags += child.name+"|";
					if(child.childTags) childTags += child.childTags;
					
				}
			}
			
			if(childTags.search(regexp.divToPElements) == -1)
			{
				nodesToScore.push(node);
			}
			
			node.textContent = textContent;
			node.childTags = childTags;
			
			
			
			return textContent || '';
		}
		
		function setParentChildLink(node)
		{
			if(node && node.children)
			for(var i = node.children.length; i--;)
			{
				var child = node.children[i];
				child.parent = node;
				setParentChildLink(child);
			}
		}
		
		function getClassWeight(node)
		{
			var weight = 0,
				attrs = node.attribs,
				className = attrs ? attrs['class'] : null,
				id = attrs ? attrs.id : null;
				
			if(className)
			{
				if(className.search(regexp.negative) != -1)
				{
					weight -= 25;
				}
				else
				if(className.search(regexp.positive) != -1)
				{
					weight += 25;
				}
			}
			
			if(id)
			{
				if(id.search(regexp.negative) != -1)
				{
					weight -= 25;
				}
				else
				if(id.search(regexp.positive) != -1)
				{
					weight += 25;
				}
			}
			
			return weight;
		}
		
		
		function initArticleCandidate(node)
		{
			node.readly = true;
			node.score = (tagScores[node.name] || 0)+getClassWeight(node);
			return node;	
		}
		
		function getInnerLinks(node,links)
		{
			if(node.name == 'a')
			{
				links.push(node);
			}
			else
			if(node.children)
			for(var i = node.children.length; i--;) getInnerLinks(node.children[i],links);
		}
		
		function getLinkDensity(node)
		{
			if(node.linkDensity != undefined) return node.linkDensity;
			
			var links = [],
				textLength = node.textContent.length,
				linkLength = 0;
				
			getInnerLinks(node,links);
			
			for(var i = links.length; i--;)
			{
				//could be wrapping an image.
				linkLength += (links[i].textContent || '').length;
			}
			
			
			
			return node.linkDensity = (linkLength/textLength || 0);
		}
		
		function getVideo(candidates, child, distance)
		{
			return { type: 'video',
							 child: child, 
							 score: 100 - distance,
							 distance: distance };
		}
		
		this.addMetadataHandler({
			addData: function(candidates, child, distanceFromArticle, ops)
			{
				if(child.name == 'img')
				{
					var data = getImageData(child, distanceFromArticle, ops);         
					                                                                 
					
					if(!data || data.score < -20) return null;     
					
					candidates.push(data);
				}
				
				return null;
			}
		});
		
		this.addMetadataHandler({
			addData: function(candidates, child, distanceFromArticle, ops)
			{
				if(child.name == 'object')
				{
					var cattrs = child.attribs;
					
					if(cattrs && (cattrs.src || '').search(regexp.positiveVideoCheck) > -1)
					{
						return candidates.push(getVideo(child, distanceFromArticle));
					}
					else
					if(child.children)
					{
						for(var j = child.children.length; j--;)
						{
							var param = child.children[j];

							//ick
							if(cattrs && (cattrs.value || '').search(regexp.positiveVideoCheck) > -1)
							{
								candidates.push(getVideo(child, distanceFromArticle));
							}
						}
					}
				}
				
				return null;
			}
		});
		
		this.addMetadataHandler({
			addData: function(candidates, child, distanceFromArticle, ops)
			{
				if(child.name == 'iframe')
				{
					var cattrs = child.attribs;
				
					if(cattrs && (cattrs.src || '').search(regexp.positiveVideoCheck) > -1)
					{
						return getVideo(candidates,child,distanceFromArticle);
					}
					
					return null;
				}
			}
		});
		
		function buildIngredient(child, parts)
		{
			if(!parts) parts = [];
			var children = child.children || [];
			
			if(child.data)
			{
				parts.error(child.data);
			}
			
			for(var i = 0, n = children.length; i < n; i++)
			{
				buildIngredient(children[i], parts);
			}
			
			return parts;
		}
		
		function findIngredientListCandidates(child, level)
		{
			if(!child) return []
			
			if(!level) level = 0;
			
			var children = child.children || [],
				candidates = [];
			
			
			for(var i = children.length; i--;)
			{
				var c = children[i];
				
				c.usedIngredients = true;
				
				
				candidates.push(buildIngredient(c));
			}
			
			
			child.usedIngredients = true;
			
			if(level < 3) candidates = candidates.concat(findIngredientListCandidates(child.parent, level + 1));
			
			return candidates;
		}
		
		
		var test = false;
		/*this.addMetadataHandler({
			addData: function(candidates, child)
			{
				
				if(test) return;
				
				
				// consol
				if(child.usedIngredients) return;
				// if(candidates.foundIngredients) return;
				
				if(child.data && child.data.match(/^(\W)*(\d{1,2}[\d\/]*)+\s/))
				{
					
					if(child.data.indexOf('30') > -1) return;
					test = true;
					
					
					var candidates = findIngredientListCandidates(child.parent),
						bestCandidate,
						score = 0;
						
						var mesurementRegexp = /cup|teaspoon|gallon|tablesoon|oz/,
							termsRegexp = /pinch/
						
						
					
					for(var i = candidates.length; i--;)
					{
						var candidate = candidates[i];
						
						for(var j = candidate.length; j--;)
						{
							var item = candidate[j];
							
							
							
						}
						
						if(!bestCandidate || candidate.length > bestCandidate.length)
						{
							bestCandidate = candidate;
						}
					}
					
					
					candidates.push({type: 'ingredients', data: bestCandidate });
					
					// candidates.foundIngredients = true;
				}
				
				//if(content.match(/()/))
				//{
				//}
			}
		})*/
		
		function findMediaCandidates(node, distanceFromArticle, down, ops, candidates)
		{

			var target = !down ? node.parent : node;
			
			if(!candidates) candidates = [];
			if(!target) return candidates;	

			for(var i = 0, n = target.children.length; i < n; i++)
			{      

				distanceFromArticle++;

				var child = target.children[i],
					cattrs = child.attribs,
					metadata;
					
				// if(child.type != 'tag') continue;
				
				for(var j = metadataHandlers.length; j--;)
				{
					var handler = metadataHandlers[j];
					
					if(metadata = handler.addData(candidates, child, distanceFromArticle, ops))
					{
						break;
					}
				}
				
				
				if(!metadata && child.children && child != node)
				{               
					if((child.textContent || '').split(regexp.punctuation).length > 5)
					{
						// distanceFromArticle = 0;
					}             
					findMediaCandidates(child, distanceFromArticle, true, ops, candidates);
				}
			}


			//of we have to go UP, then we're getting into content we really REALLY don't want.
			if(!down) findMediaCandidates(target, distanceFromArticle, false, ops, candidates);

			return candidates;
		}
		
		
		function getArticleMedia(articleParts,ops)
		{
			var richest = findRichestNode(articleParts);
			
			return richest ? findMediaCandidates(richest.node, 0, false, ops) : [];
		}
		
		function fixUrl(url,ops)
		{
			if(url && url.substr(0,8).indexOf('://') == -1)
			{
				var dir = ''; 
				
				if(url.substr(0,1) != '/') dir = ops.location.pathdir;
				url = 'http://'+((ops.location.host+'/'+dir+'/')+url).replace(/\/+/g,'/');
			}
			
			return (url || '').replace(/&amp;/g,'&');;
		} 
		
		function cleanArticle(parts)
		{
			if(parts)
			{
				for(var i = parts.length; i--;)
				{
					var node = parts[i];
					
					if(node.name && node.name.search(regexp.cleanRemoveElements) > -1)
					{            
						parts.splice(i,1);         
						continue;
					}  
					
					cleanArticle(node.children);
				}
			}
		}
		
		
		
		function rebuildArticle(parts,ops,metadata)
		{
			var article = '';
			
			if(parts)
			{
				for(var i = parts.length; i--;)
				{
					var node = parts[i],
						innerText = (rebuildArticle(node.children,ops,metadata) || node.data || ' ');

					if(node.name == 'a')
					{
						var attribs = node.attribs;
						if(attribs)
						{
							var href = attribs.href;
							// metadata.push({link:})
							
							if(node.children)
							{
								var child = node.children[0];
								
								if(child && child.name != 'img')
								{
									metadata.push({type:'url',link:fixUrl(href,ops),match:innerText.replace(/^\s+|\s+$/gi,'')});
								}
								else
								{
									innerText = '';
								}
							}
						}
					}
					else
					if(node.name && keepNodes[node.name])	
					// if(node.name && node.name.match(/^[a-z0-9]+$/i) && node.name.length < 3 && innerText.match(/[^\s]/))
					{
						innerText = innerText.length > 0 ? '<'+node.name+'>'+innerText+'</'+node.name+'>' : '';
					}                
					
					
					
					article = innerText+' '+article;
				}
				return article;
			}
			
			return null;
			
		}
		
		function buildArticle(body,ops)
		{
			var nodesToScore = [],
				candidates = [];
				
			findArticleCandidates(body,nodesToScore);
			
			for(var i = nodesToScore.length; i--;)
			{
				var nodeToScore = nodesToScore[i],
					textContent = nodeToScore.textContent,
					parent = nodeToScore.parent,
					grandParent = parent ? parent.parent : null,
					score = 1;
					
				if(parent && !parent.readly)
				{
					candidates.push(initArticleCandidate(parent));
				}
				
				if(grandParent && !grandParent.readly)
				{
					candidates.push(initArticleCandidate(grandParent));
				}
					
				if(textContent.length < 25) continue;
				
				
				//add points for punctiation such as periods, and commas
				score += textContent.split(regexp.punctuation).length;
				
				//For every 100 characters in this paragraph, add another point. Up to 3 points.
				score += Math.min(Math.floor(textContent.length/100),3);
				
				if(parent) parent.score += score;
				if(grandParent)	grandParent.score += score/2;
				
			}
			
			var topCandidate = null;
			
			
			for(var i = candidates.length; i--;)
			{
				var candidate = candidates[i];
				
				if(candidate.textContent.length < 25) continue;
				
				//scale back the article based on the link density. more links = bad
				candidate.score = candidate.score * (1-getLinkDensity(candidate));
				
				//NOTE: the commented out code is here for reference. The reason WHY it's commented out is because
				//candidates can be in other parts of the page, and so can images with them. We do NOT WANT IMAGES
				//that don't relate to the article!!
				// var finalScore =  candidate.score;// * Math.max(.3,_setImageScore(candidate));
				
				if(!topCandidate || candidate.score > topCandidate.score)
				{
					topCandidate = candidate;
				}
			}
			
			
			// topCandidate = topCandidate || body;
			
			if(!topCandidate) return {article:[],media:[]};
			
			var topCandidateClass = topCandidate.attribs ? topCandidate.attribs['class'] : '';
			/**
		     * Now that we have the top candidate, look through its siblings for content that might also be related.
		     * Things like preambles, content split by ads that we removed, etc.
		    **/
		
			var siblingScoreThreshold = Math.min(10,topCandidate.score*0.5);
			
			var siblings = topCandidate.parent ? topCandidate.parent.children : [],
				contentChildren = [],
				textContent = '';
				
			for(var i = 0, n = siblings.length; i < n; i++)
			{
				var sib = siblings[i],
					contentBonus = 0,
					append = sib == topCandidate,
					attrs = sib.attribs,
					className = attrs ? attrs['class'] : null;
					
					
				if(className == topCandidateClass && topCandidateClass != '')
				{
					contentBonus += topCandidate.score * 0.2;
				}
				
				if(sib.score+contentBonus >= siblingScoreThreshold)
				{
					append = true;
				}
				else
				if(sib.name == 'p')
				{
					var linkDensity = getLinkDensity(sib),
						c = sib.textContent || '',
						nodeLength = c.length;
						
					append = c.split(regexp.punctuation).length > 3 || 
							 (nodeLength > 80 && linkDensity < 0.3) ||
							 (nodeLength < 80 && linkDensity === 0 && c.search(/\.( |$)/) !== -1);
				}
				
				if(append)
				{
					contentChildren.push(sib);
					textContent += sib.textContent+' ';
				}
			}
			
			textContent = textContent.replace(/\s/g,' ');//clean up the whitespace.
			
			if(textContent.length < 25 && false)
			{
				if(SKIP_UNLIKELY_CANDIDATES)
				{
					SKIP_UNLIKELY_CANDIDATES = false;
					return buildArticle(body, ops);
				}
			}
			
			//let's find the images!
			var media = getArticleMedia(contentChildren,ops);
			//get article media here :3
			
			
			var metadata = [];
			                              
			//remove shit that we don't want.
			cleanArticle(contentChildren);
			
			//rebuild the article text now ;)
			var articleText = rebuildArticle(contentChildren,ops,metadata).replace(/(\s|&nbsp;)+/g,' ');
			
			
			return { content: articleText, 
					 media: media, 
					 metadata: metadata, 
					 url: ops.location.url };
		}
	
}


exports.Parser = Parser;